import torch
import torch.nn as nn
from einops import rearrange, repeat
from torch import Tensor
from torch.nn import Parameter
from torch.nn import functional as F
from typing import Optional
import math
import torch.nn.init as init


class RPQEmbedding(nn.Module):
    """A simple lookup table that stores embeddings of a fixed dictionary and size using a rpq weight matrix.
    
       For padding_idx functionality, the padded embeddings are not initialized to 0.
    
    """
    __constants__ = ['num_embeddings', 'embedding_dim', 'codebook_dim', 'max_norm',
                     'norm_type', 'scale_grad_by_freq', 'sparse', 'num_codebooks']
    num_embeddings: int
    embedding_dim: int
    num_codebooks: int
    padding_idx: Optional[float]
    max_norm: Optional[float]
    norm_type: float
    scale_grad_by_freq: bool
    weight: Tensor
    sparse: bool
    num_codebooks: int
    
    def __init__(self, num_embeddings: int, embedding_dim: int, num_codebooks: int, use_subset=True,
                 padding_idx: Optional[int] = None, max_norm: Optional[float] = None, norm_type: float = 2., 
                 scale_grad_by_freq: bool = False, sparse: bool = False, 
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        if padding_idx is not None:
            if padding_idx > 0:
                assert padding_idx < self.num_embeddings, 'Padding_idx must be within num_embeddings'
            elif padding_idx < 0:
                assert padding_idx >= -self.num_embeddings, 'Padding_idx must be within num_embeddings'
                padding_idx = self.num_embeddings + padding_idx
        self.padding_idx = padding_idx
        
        self.num_codebooks = num_codebooks
        assert self.embedding_dim % num_codebooks == 0, 'embedding_dim should be divisible by num_codebooks'
        self.codebook_dim = self.embedding_dim//self.num_codebooks
        
        self.use_subset = use_subset
        self.max_norm = max_norm
        self.norm_type = norm_type
        self.scale_grad_by_freq = scale_grad_by_freq
        self.sparse = sparse

        self.rpqweight = RPQWeight(self.num_codebooks, self.codebook_dim, self.num_embeddings,
                                   device=factory_kwargs['device'], dtype=factory_kwargs['dtype'])
                                         
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.normal_(self.rpqweight.codebooks)

    def forward(self, input: Tensor) -> Tensor:
        if self.use_subset:
            return self.rpqweight(subset=input.flatten()).view(*input.shape, self.embedding_dim)
        else:
            return F.embedding(
                input, self.rpqweight(), self.padding_idx, self.max_norm, 
                self.norm_type, self.scale_grad_by_freq, sparse=self.sparse)

    def extra_repr(self) -> str:
        s = '{num_embeddings}, {embedding_dim}'
        if self.padding_idx is not None:
            s += ', padding_idx={padding_idx}'
        if self.max_norm is not None:
            s += ', max_norm={max_norm}'
        if self.norm_type != 2:
            s += ', norm_type={norm_type}'
        if self.scale_grad_by_freq is not False:
            s += ', scale_grad_by_freq={scale_grad_by_freq}'
        if self.sparse is not False:
            s += ', sparse=True'
        return s.format(**self.__dict__)


class RPQLinear(nn.Module):
    """Applies linear transformation to the incoming data."""
    __constants__ = ['in_features', 'out_features', 'num_codebooks']
    in_features: int
    out_features: int
    num_codebooks: int
    codebooks: Tensor
    
    def __init__(self, in_features: int, out_features: int, num_codebooks: int, shared_codebooks=False, 
                 split='column', bias:bool = True, device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.num_codebooks = num_codebooks
        self.split = split
        self.shared_codebooks = shared_codebooks
        
        if self.split == 'row':
            assert self.out_features % num_codebooks == 0, 'out_features should be divisible by num_codebooks'
            self.codebook_dim = self.out_features//self.num_codebooks
            self.register_buffer("codes",
                                torch.randint(high=256, size=(self.num_codebooks, self.in_features), 
                                            dtype=torch.uint8, device=factory_kwargs['device']))
        elif self.split == 'column':
            assert self.in_features % num_codebooks == 0, 'in_features should be divisible by num_codebooks'
            self.codebook_dim = self.in_features//self.num_codebooks
            self.register_buffer("codes",
                                torch.randint(high=256, size=(self.num_codebooks, self.out_features), 
                                            dtype=torch.uint8, device=factory_kwargs['device']))
        if not self.shared_codebooks:
            self.codebooks = Parameter(torch.empty(self.num_codebooks, 256, self.codebook_dim, **factory_kwargs))
        
        if bias:
            self.bias = Parameter(torch.empty(self.out_features, **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not self.shared_codebooks:
            init.kaiming_uniform_(self.codebooks, a=math.sqrt(5))
        if self.bias is not None:
            # manually get fan_in
            fan_in = self.in_features
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)
            
    def expand(self, codes, codebooks):
        codes_expand = repeat(codes, 'h c -> h c d', d = self.codebook_dim)
        return codebooks.gather(dim=1, index=codes_expand.long())
        
    def get_weight(self) -> Tensor:
        return rearrange(self.expand(self.codes, self.codebooks), 
                         'h c d -> c (h d)')

    def forward(self, input: Tensor) -> Tensor:
        return F.linear(input, self.get_weight(), self.bias)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, num_codebooks={}, split={}'.format(
            self.in_features, self.out_features, self.num_codebooks, self.split
        )
    

class RPQWeight(nn.Module):
    """A standalone layer that initializes an RPQ weight matrix from a given full weight matrix."""
    __constants__ = ['num_codebooks', 'codebook_dim', 'num_vectors']
    num_codebooks: int
    codebook_dim: int
    num_vectors: int
    codebooks: Tensor
    
    def __init__(self, num_codebooks: int, codebook_dim: int, num_vectors: int,  shared_codebooks=False, 
                 split: str = 'column', device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_codebooks = num_codebooks
        self.codebook_dim = codebook_dim
        self.num_vectors = num_vectors
        self.shared_codebooks = shared_codebooks
        self.split = split
        if (self.split != 'row') and (self.split != 'column'):
            raise ValueError(f"split must be either 'row' or 'column', got {self.split}")
        
        self.register_buffer("codes",
                             torch.randint(high=256, size=(self.num_codebooks, self.num_vectors), 
                                           dtype=torch.uint8, device=factory_kwargs['device']))
        if not self.shared_codebooks:
            self.codebooks = Parameter(torch.empty(self.num_codebooks, 256, 
                                                self.codebook_dim, **factory_kwargs))            
        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not self.shared_codebooks:
            init.kaiming_normal_(self.codebooks)

    def expand(self, codes, codebooks, subset):
        codes_expand = repeat(codes[:, subset], 'h c -> h c d', d=self.codebook_dim)
        weight = codebooks.gather(dim=1, index=codes_expand.long())
        weight = rearrange(weight, 'h c d -> c (h d)')
        return weight

    def forward(self, subset=slice(None)) -> Tensor:
        return self.expand(self.codes, self.codebooks, subset)

    def extra_repr(self) -> str:
        s = 'num_codebooks={}, codebook_dim={}, num_vectors={}'.format(
            self.num_codebooks, self.codebook_dim, self.num_vectors)
        return s



import math

def model_size(model):
    """Returns the size of the model in MB."""
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()
    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()

    size_all_mb = (param_size + buffer_size) / 1024**2
    print('model size: {:.3f}MB'.format(size_all_mb))

def calc_head_expansion(vocab_size, hidden_dim, num_heads, min_subvector_size=32):
    expansion = 2**round(math.log(vocab_size/hidden_dim, 2))
    subvector_size = max(min_subvector_size, hidden_dim//(expansion*num_heads))
    return hidden_dim//subvector_size


from setuptools import setup, find_packages

setup(
  name = 'RPQ-pytorch',
  packages = find_packages(exclude=[]),
  version = '0.0.32',
  license='MIT',
  description = 'Reverse Product Quantization (RPQ) of weights to reduce static memory usage.',
  author = 'Ali Kore',
  author_email = 'akore654@gmail.com',
  long_description=open('README.md', 'r').read(),
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/a-kore/RPQ-pytorch',
  keywords = [
    'artificial intelligence',
    'AI',
    'machine learning',
    'deep learning',
    'pytorch',
    'quantization',
    'product quantization',
    'reverse product quantization',
    'memory reduction',
  ],
  install_requires=[
    'torch>=1.6',
    'einops>=0.6',
    'transformers>=4.0',
    'vit-pytorch>=0.40',
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.10',
  ],
)


"""
Implementation of ViT model with standard linear modules replaced with their RPQ counterparts.

This is a modified version of the ViT model from the https://github.com/lucidrains/vit-pytorch repository.
"""
import torch
from torch import nn
from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from torch.nn.modules.utils import _pair
from rpq.nn import RPQLinear
import math


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn
    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)

class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, num_codebooks, rpq_codebooks: nn.ParameterList, dropout = 0., expand_heads=False):
        super().__init__()

        if expand_heads:
            expansion = (hidden_dim//dim)
        else:
            expansion = 1
        self.fc1 = RPQLinear(dim, hidden_dim, num_codebooks*expansion, shared_codebooks=True)
        self.fc1.codebooks = rpq_codebooks[2]
        self.fc2 = RPQLinear(hidden_dim, dim, num_codebooks*expansion, shared_codebooks=True)
        self.fc2.codebooks = rpq_codebooks[3]

        self.net = nn.Sequential(
            self.fc1,
            nn.GELU(),
            nn.Dropout(dropout),
            self.fc2,
            nn.Dropout(dropout)
        )
    def forward(self, x):
        return self.net(x)

class Attention(nn.Module):
    def __init__(self, dim, rpq_codebooks: nn.ParameterList, heads = 8, dim_head = 64, dropout = 0.):
        super().__init__()
        inner_dim = dim_head *  heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim = -1)
        self.dropout = nn.Dropout(dropout)

        self.to_qkv = RPQLinear(dim, inner_dim * 3, heads, bias = False, shared_codebooks=True)
        self.to_qkv.codebooks = rpq_codebooks[0]

        if project_out:
            self.out_proj = RPQLinear(inner_dim, dim, heads, shared_codebooks=True)
            self.out_proj.codebooks = rpq_codebooks[1]
            self.to_out = nn.Sequential(
                self.out_proj,
                nn.Dropout(dropout)
            )
        else:
            self.to_out = nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim = -1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)

class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout = 0., expand_heads=False):
        super().__init__()
        if expand_heads:
            expansion = (mlp_dim//dim)
        else:
            expansion = 1
        self.rpq_codebooks = nn.ParameterList(
            [torch.empty(heads, 256, dim // heads),
            torch.empty(heads, 256, dim // heads),
            torch.empty(heads*expansion, 256, mlp_dim // heads),
            torch.empty(heads*expansion, 256, mlp_dim // heads)]
        )
        # init all codebooks using kaiming_uniform_
        for codebook in self.rpq_codebooks:
            nn.init.kaiming_uniform_(codebook, a=math.sqrt(5))

        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, self.rpq_codebooks, heads = heads, dim_head = dim_head, dropout = dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, heads*(mlp_dim//dim), self.rpq_codebooks, dropout = dropout, expand_heads=expand_heads))
            ]))
    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return x

class RPQViT(nn.Module):
    def __init__(self, 
                 *, 
                 image_size, 
                 patch_size, 
                 num_classes, 
                 dim, 
                 depth, 
                 heads, 
                 mlp_dim, 
                 pool = 'cls', 
                 channels = 3, 
                 dim_head = 64, 
                 dropout = 0., 
                 emb_dropout = 0.,
                 expand_heads = False
                 ):
        super().__init__()
        image_height, image_width = _pair(image_size)
        patch_height, patch_width = _pair(patch_size)

        assert image_height % patch_height == 0 and image_width % patch_width == 0, 'Image dimensions must be divisible by the patch size.'

        num_patches = (image_height // patch_height) * (image_width // patch_width)
        patch_dim = channels * patch_height * patch_width
        assert pool in {'cls', 'mean'}, 'pool type must be either cls (cls token) or mean (mean pooling)'

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = patch_height, p2 = patch_width),
            nn.Linear(patch_dim, dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, heads, dim_head, mlp_dim, dropout, expand_heads)

        self.pool = pool
        self.to_latent = nn.Identity()

        self.mlp_head = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, num_classes)
        )

    def forward(self, img):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, '1 1 d -> b 1 d', b = b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        x = x.mean(dim = 1) if self.pool == 'mean' else x[:, 0]

        x = self.to_latent(x)
        return self.mlp_head(x)






